from typing import AsyncGenerator, List
import logging
from aiconsole.code_interpreters.language_map import language_map
from aiconsole.gpt.consts import GPTMode
from aiconsole.gpt.gpt_executor import GPTExecutor
from aiconsole.gpt.parse_partial_json import parse_partial_json
from aiconsole.gpt.request import GPTRequest
from aiconsole.gpt.types import CLEAR_STR, GPTMessage

log = logging.getLogger(__name__)

function_schema = {
  "name": "execute",
  "description":
  "Executes code on the user's machine, **in the users local environment**, and returns the output",
  "parameters": {
    "type": "object",
    "properties": {
      "language": {
        "type": "string",
        "description":
        "The programming language (required parameter to the `execute` function)",
        "enum": [language for language in language_map.keys()]
      },
      "code": {
        "type": "string",
        "description": "The code to execute (required)"
      }
    },
    "required": ["language", "code"]
  },
}


async def coding_llm(messages: List[GPTMessage], system_message: str, gpt_mode: GPTMode, min_tokens: int, preffered_tokens: int) -> AsyncGenerator[str, None]:
    
    executor = GPTExecutor()
    
    language = None
    code = ""

    async for chunk in executor.execute(GPTRequest(system_message=system_message, gpt_mode=gpt_mode, messages = messages,  functions = [function_schema], min_tokens = min_tokens, preferred_tokens= preffered_tokens)):
        if chunk == CLEAR_STR:
            yield CLEAR_STR
            continue

        if ('choices' not in chunk or len(chunk['choices']) == 0):
            continue

        delta = chunk["choices"][0]["delta"]

        if "content" in delta and delta["content"]:
            yield delta["content"]

        function_call = executor.partial_response.choices[0].message.function_call

        if function_call and function_call.arguments:
            # This can now be both a string and a json object
            arguments = parse_partial_json(function_call.arguments)
            languages = language_map.keys()
            
            # We need to handle incorrect OpenAI responses, sometmes arguments is a string containing the code
            if isinstance(arguments, str):
                if language is None and function_call.name in languages:
                    language = function_call.name
                    yield f'<<<< START CODE ({language}) >>>>'

                if arguments and not arguments.startswith("{"):
                    if language is None:
                        language = "python"
                        yield f'<<<< START CODE ({language}) >>>>'

                    code_delta = arguments[len(code):]
                    code = arguments
                    
                    if code_delta:
                        yield code_delta

            else:
                if language is None and arguments and "language" in arguments and arguments["language"] in languages:
                    language = arguments["language"]
                    yield f'<<<< START CODE ({language}) >>>>'

                if language is not None and arguments and "code" in arguments:
                    code_delta = arguments["code"][len(code):]
                    code = arguments["code"]
                    
                    if code_delta:
                        yield code_delta


    # If the code starts with a !, that means a shell command
    if language == "python" and code.startswith("!"):
        language = "shell"
        # Clear the message and resend it
        yield CLEAR_STR
        yield f'<<<< START CODE ({language}) >>>>'
        yield code[1:]
    
    if language:
        yield f"<<<< END CODE >>>>"
        language = None
